@@ -14,10 +14,17 @@ use Graph::AdjacencyMap qw(:flags :fields);
 
 use vars qw($VERSION);
 
-$VERSION = '0.78';
+$VERSION = '0.91';
 
 require 5.006; # Weak references are absolutely required.
 
+my $can_deep_copy_Storable =
+    eval 'require Storable; require B::Deparse; $Storable::VERSION >= 2.05 && $B::Deparse::VERSION >= 0.61' && !$@;
+
+sub _can_deep_copy_Storable () {
+    return $can_deep_copy_Storable;
+}
+
 use Graph::AdjacencyMap::Heavy;
 use Graph::AdjacencyMap::Light;
 use Graph::AdjacencyMap::Vertex;
@@ -28,8 +35,11 @@ use Graph::MSTHeapElem;
 use Graph::SPTHeapElem;
 use Graph::Undirected;
 
-use Heap::Fibonacci;
+use Heap071::Fibonacci;
 use List::Util qw(shuffle first);
+use Scalar::Util qw(weaken);
+
+use Safe;  # For deep_copy().
 
 sub _F () { 0 } # Flags.
 sub _G () { 1 } # Generation.
@@ -37,6 +47,20 @@ sub _V () { 2 } # Vertices.
 sub _E () { 3 } # Edges.
 sub _A () { 4 } # Attributes.
 sub _U () { 5 } # Union-Find.
+sub _S () { 6 } # Successors.
+sub _P () { 7 } # Predecessors.
+
+my $Inf;
+
+BEGIN {
+    local $SIG{FPE}; 
+    eval { $Inf = exp(999) } ||
+	eval { $Inf = 9**9**9 } ||
+	    eval { $Inf = 1e+999 } ||
+		{ $Inf = 1e+99 };  # Close enough for most practical purposes.
+}
+
+sub Infinity () { $Inf }
 
 # Graphs are blessed array references.
 # - The first element contains the flags.
@@ -65,8 +89,8 @@ sub _COMPAT02 () { 0x00000001 }
 
 sub stringify {
     my $g = shift;
-    my $o = $g->is_undirected;
-    my $e = $o ? '=' : '-';
+    my $u = $g->is_undirected;
+    my $e = $u ? '=' : '-';
     my @e =
 	map {
 	    my @v =
@@ -74,7 +98,7 @@ sub stringify {
 		    ref($_) eq 'ARRAY' ? "[" . join(" ", @$_) . "]" : "$_"
 		}
 	    @$_;
-	    join($e, $o ? sort { "$a" cmp "$b" } @v : @v) } $g->edges05;
+	    join($e, $u ? sort { "$a" cmp "$b" } @v : @v) } $g->edges05;
     my @s = sort { "$a" cmp "$b" } @e;
     push @s, sort { "$a" cmp "$b" } $g->isolated_vertices;
     join(",", @s);
@@ -639,7 +663,7 @@ sub add_edge_by_id {
     $g->expect_multiedged;
     my $id = pop;
     my @e = $g->_add_edge( @_ );
-    $g->[ _E ]->set_path( @e, $id );
+    $g->[ _E ]->set_path_by_multi_id( @e, $id );
     $g->[ _G ]++;
     $g->_union_find_add_edge( @e ) if $g->has_union_find;
     return $g;
@@ -747,108 +771,59 @@ sub _edges_at {
     return wantarray ? @e : $en;
 }
 
-sub _edges_from {
+sub _edges {
     my $g = shift;
+    my $n = pop;
+    my $i = $n == _S ? 0 : -1;  # _edges_from() or _edges_to()
     my $V = $g->[ _V ];
     my $E = $g->[ _E ];
-    my @e;
-    my $o = $E->[ _f ] & _UNORD;
-    my $en = 0;
-    my %ev;
-    my $h = $V->[_f ] & _HYPER;
-    for my $v ( $h ? $g->vertices_at( @_ ) : @_ ) {
-	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
-	next unless defined $vi;
+    my $N = $g->[ $n ];
+    my $h = $V->[ _f ] & _HYPER;
+    unless (defined $N && $N->[ 0 ] == $g->[ _G ]) {
+	$g->[ $n ]->[ 1 ] = { };
+	$N = $g->[ $n ];
+	my $u = $E->[ _f ] & _UNORD;
 	my $Ei = $E->_ids;
-	if (wantarray) {
-	    if ($o) {
-		while (my ($ei, $ev) = each %{ $Ei }) {
-		    next unless @$ev;
-		    push @e, [ $ei, $ev ]
-			if ($ev->[0] == $vi || $ev->[-1] == $vi) && !$ev{$ei}++;
-		}
+	while (my ($ei, $ev) = each %{ $Ei }) {
+	    next unless @$ev;
+	    my $e = [ $ei, $ev ];
+	    if ($u) {
+		push @{ $N->[ 1 ]->{ $ev->[ 0] } }, $e;
+		push @{ $N->[ 1 ]->{ $ev->[-1] } }, $e;
 	    } else {
-		while (my ($ei, $ev) = each %{ $Ei }) {
-		    next unless @$ev;
-		    push @e, [ $ei, $ev ]
-			if $ev->[0] == $vi && !$ev{$ei}++;
-		}
-	    }
-	} else {
-	    if ($o) {
-		while (my ($ei, $ev) = each %{ $Ei }) {
-		    next unless @$ev;
-		    $en++ if ($ev->[0] == $vi || $ev->[-1] == $vi);
-		}
-	    } else {
-		while (my ($ei, $ev) = each %{ $Ei }) {
-		    next unless @$ev;
-		    $en++ if $ev->[0] == $vi;
-		}
+		my $e = [ $ei, $ev ];
+		push @{ $N->[ 1 ]->{ $ev->[$i] } }, $e;
 	    }
 	}
+	$N->[ 0 ] = $g->[ _G ];
+    }
+    my @e;
+    my @at = $h ? $g->vertices_at( @_ ) : @_;
+    my %at; @at{@at} = ();
+    for my $v ( @at ) {
+	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
+	next unless defined $vi && exists $N->[ 1 ]->{ $vi };
+	push @e, @{ $N->[ 1 ]->{ $vi } };
     }
     if (wantarray && $g->is_undirected) {
 	my @i = map { $V->_get_path_id( $_ ) } @_;
 	for my $e ( @e ) {
-	    unless ( $e->[ 1 ]->[ 0 ] == $i[ 0 ] ) { # @todo
+	    unless ( $e->[ 1 ]->[ $i ] == $i[ $i ] ) {
 		$e = [ $e->[ 0 ], [ reverse @{ $e->[ 1 ] } ] ];
 	    }
 	}
     }
-    return wantarray ? @e : $en;
+    return @e;
+}
+
+sub _edges_from {
+    push @_, _S;
+    goto &_edges;
 }
 
 sub _edges_to {
-    my $g = shift;
-    my $V = $g->[ _V ];
-    my $E = $g->[ _E ];
-    my @e;
-    my $o = $E->[ _f ] & _UNORD;
-    my $en = 0;
-    my %ev;
-    my $h = $V->[_f ] & _HYPER;
-    for my $v ( $h ? $g->vertices_at( @_ ) : @_ ) {
-	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
-	next unless defined $vi;
-	my $Ei = $E->_ids;
-	if (wantarray) {
-	    if ($o) {
-		while (my ($ei, $ev) = each %{ $Ei }) {
-		    next unless @$ev;
-		    push @e, [ $ei, $ev ]
-			if ($ev->[-1] == $vi || $ev->[0] == $vi) && !$ev{$ei}++;
-		}
-	    } else {
-		while (my ($ei, $ev) = each %{ $Ei }) {
-		    next unless @$ev;
-		    push @e, [ $ei, $ev ]
-			if $ev->[-1] == $vi && !$ev{$ei}++;
-		}
-	    }
-	} else {
-	    if ($o) {
-		while (my ($ei, $ev) = each %{ $Ei }) {
-		    next unless @$ev;
-		    $en++ if $ev->[-1] == $vi || $ev->[0] == $vi;
-		}
-	    } else {
-		while (my ($ei, $ev) = each %{ $Ei }) {
-		    next unless @$ev;
-		    $en++ if $ev->[-1] == $vi;
-		}
-	    }
-	}
-    }
-    if (wantarray && $g->is_undirected) {
-	my @i = map { $V->_get_path_id( $_ ) } @_;
-	for my $e ( @e ) {
-	    unless ( $e->[ 1 ]->[ -1 ] == $i[ -1 ] ) { # @todo
-		$e = [ $e->[ 0 ], [ reverse @{ $e->[ 1 ] } ] ];
-	    }
-	}
-    }
-    return wantarray ? @e : $en;
+    push @_, _P;
+    goto &_edges;
 }
 
 sub _edges_id_path {
@@ -890,6 +865,58 @@ sub predecessors {
 	Graph::AdjacencyMap::_predecessors($E, $g, @_);
 }
 
+sub _all_successors {
+    my $g = shift;
+    my @init = @_;
+    my %todo;
+    @todo{@init} = @init;
+    my %seen;
+    while (keys %todo) {
+	my @todo = keys %todo;
+	for my $t (@todo) {
+	    delete $todo{$t};
+	    $seen{$t} = $t;
+	    for my $s ($g->successors($t)) {
+		$todo{$s}++ unless exists $seen{$s};
+	    }
+	}
+    }
+    delete @seen{@init};
+    return values %seen;
+}
+
+sub all_successors {
+    my $g = shift;
+    $g->expect_directed;
+    return $g->_all_successors(@_);
+}
+
+sub _all_predecessors {
+    my $g = shift;
+    my @init = @_;
+    my %todo;
+    @todo{@init} = @init;
+    my %seen;
+    while (keys %todo) {
+	my @todo = keys %todo;
+	for my $t (@todo) {
+	    delete $todo{$t};
+	    $seen{$t} = $t;
+	    for my $p ($g->predecessors($t)) {
+		$todo{$p}++ unless exists $seen{$p};
+	    }
+	}
+    }
+    delete @seen{@init};
+    return values %seen;
+}
+
+sub all_predecessors {
+    my $g = shift;
+    $g->expect_directed;
+    return $g->_all_predecessors(@_);
+}
+
 sub neighbours {
     my $g = shift;
     my $V  = $g->[ _V ];
@@ -903,6 +930,25 @@ sub neighbours {
 
 *neighbors = \&neighbours;
 
+sub all_neighbours {
+    my $g = shift;
+    my @init = @_;
+    my %n;
+    my @p = $g->_all_predecessors(@init);
+    my @s = $g->_all_successors(@init);
+    @n{@p} = @p;
+    @n{@s} = @s;
+    delete @n{@init};
+    return values %n;
+}
+
+*all_neighbors = \&all_neighbours;
+
+sub all_reachable {
+    my $g = shift;
+    $g->directed ? $g->successors(@_) : $g->neighbors(@_);
+}
+
 sub delete_edge {
     my $g = shift;
     my @i = $g->_vertex_ids( @_ );
@@ -964,7 +1010,7 @@ sub delete_edges {
 sub _in_degree {
     my $g = shift;
     return undef unless @_ && $g->has_vertex( @_ );
-    my $in =  $g->is_undirected && $g->is_self_loop_vertex( @_ ) ? 1 : 0;
+    my $in = 0;
     $in += $g->get_edge_count( @$_ ) for $g->edges_to( @_ );
     return $in;
 }
@@ -977,7 +1023,7 @@ sub in_degree {
 sub _out_degree {
     my $g = shift;
     return undef unless @_ && $g->has_vertex( @_ );
-    my $out =  $g->is_undirected && $g->is_self_loop_vertex( @_ ) ? 1 : 0;
+    my $out = 0;
     $out += $g->get_edge_count( @$_ ) for $g->edges_from( @_ );
     return $out;
 }
@@ -999,14 +1045,12 @@ sub degree {
     my $g = shift;
     if (@_) {
 	$g->_total_degree( @_ );
+    } elsif ($g->is_undirected) {
+	my $total = 0;
+	$total += $g->_total_degree( $_ ) for $g->vertices05;
+	return $total;
     } else {
-	if ($g->is_undirected) {
-	    my $total = 0;
-	    $total += $g->_total_degree( $_ ) for $g->vertices05;
-	    return $total;
-	} else {
-	    return 0;
-	}
+	return 0;
     }
 }
 
@@ -1076,7 +1120,7 @@ sub is_self_loop_vertex {
     my $g = shift;
     return 0 unless @_;
     for my $s ( $g->successors( @_ ) ) {
-	return 1 if $s eq $_[0]; # @todo: hypervertices
+	return 1 if $s eq $_[0]; # @todo: multiedges, hypervertices
     }
     return 0;
 }
@@ -1593,6 +1637,7 @@ sub in_edges {
 sub add_vertices {
     my $g = shift;
     $g->add_vertex( $_ ) for @_;
+    return $g;
 }
 
 sub add_edges {
@@ -1611,6 +1656,7 @@ sub add_edges {
 	    }
 	}
     }
+    return $g;
 }
 
 ###
@@ -1620,18 +1666,36 @@ sub add_edges {
 sub copy {
     my $g = shift;
     my %opt = _get_options( \@_ );
-    
-    my $c = (ref $g)->new(directed => $g->directed ? 1 : 0,
-			  compat02 => $g->compat02 ? 1 : 0);
+
+    my $c =
+	(ref $g)->new(map { $_ => $g->$_ ? 1 : 0 }
+		      qw(directed
+			 compat02
+			 refvertexed
+			 hypervertexed
+			 countvertexed
+			 multivertexed
+			 hyperedged
+			 countedged
+			 multiedged
+			 omniedged));
     for my $v ($g->isolated_vertices) { $c->add_vertex($v) }
     for my $e ($g->edges05)           { $c->add_edge(@$e)  }
+
     return $c;
 }
 
 *copy_graph = \&copy;
 
-sub deep_copy {
-    require Data::Dumper;
+sub _deep_copy_Storable {
+    my $g = shift;
+    my $safe = new Safe;
+    local $Storable::Deparse = 1;
+    local $Storable::Eval = sub { $safe->reval($_[0]) };
+    return Storable::thaw(Storable::freeze($g));
+}
+
+sub _deep_copy_DataDumper {
     my $g = shift;
     my $d = Data::Dumper->new([$g]);
     use vars qw($VAR1);
@@ -1640,6 +1704,14 @@ sub deep_copy {
     eval $d->Dump;
 }
 
+sub deep_copy {
+    if (_can_deep_copy_Storable()) {
+	return _deep_copy_Storable(@_);
+    } else {
+	return _deep_copy_DataDumper(@_);
+    }
+}
+
 *deep_copy_graph = \&deep_copy;
 
 sub transpose_edge {
@@ -2253,7 +2325,7 @@ sub _heap_walk {
     my ($g, $h, $add, $etc) = splice @_, 0, 4; # Leave %opt in @_.
 
     my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) = $g->_root_opt(@_);
-    my $HF = Heap::Fibonacci->new;
+    my $HF = Heap071::Fibonacci->new;
 
     while (defined $r) {
 	# print "r = $r\n";
@@ -2520,68 +2592,131 @@ sub directed_copy {
 
 *directed_copy_graph = \&directed_copy;
 
+###
+# Cache or not.
+#
+
+my %_cache_type =
+    (
+     'connectivity'        => '_ccc',
+     'strong_connectivity' => '_scc',
+     'biconnectivity'      => '_bcc',
+     'SPT_Dijkstra'        => '_spt_di',
+     'SPT_Bellman_Ford'    => '_spt_bf',
+    );
+
+sub _check_cache {
+    my ($g, $type, $code) = splice @_, 0, 3;
+    my $c = $_cache_type{$type};
+    if (defined $c) {
+	my $a = $g->get_graph_attribute($c);
+	unless (defined $a && $a->[ 0 ] == $g->[ _G ]) {
+	    $a->[ 0 ] = $g->[ _G ];
+	    $a->[ 1 ] = $code->( $g, @_ );
+	    $g->set_graph_attribute($c, $a);
+	}
+	return $a->[ 1 ];
+    } else {
+	Carp::croak("Graph: unknown cache type '$type'");
+    }
+}
+
+sub _clear_cache {
+    my ($g, $type) = @_;
+    my $c = $_cache_type{$type};
+    if (defined $c) {
+	$g->delete_graph_attribute($c);
+    } else {
+	Carp::croak("Graph: unknown cache type '$type'");
+    }
+}
+
+sub connectivity_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'connectivity');
+}
+
+sub strong_connectivity_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'strong_connectivity');
+}
+
+sub biconnectivity_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'biconnectivity');
+}
+
+sub SPT_Dijkstra_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'SPT_Dijkstra');
+    $g->delete_graph_attribute('SPT_Dijkstra_first_root');
+}
+
+sub SPT_Bellman_Ford_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'SPT_Bellman_Ford');
+}
+
 ###
 # Connected components.
 #
 
+sub _connected_components_compute {
+    my $g = shift;
+    my %cce;
+    my %cci;
+    my $cc = 0;
+    if ($g->has_union_find) {
+	my $UF = $g->_get_union_find();
+	my $V  = $g->[ _V ];
+	my %icce; # Isolated vertices.
+	my %icci;
+	my $icc = 0;
+	for my $v ( $g->unique_vertices ) {
+	    $cc = $UF->find( $V->_get_path_id( $v ) );
+	    if (defined $cc) {
+		$cce{ $v } = $cc;
+		push @{ $cci{ $cc } }, $v;
+	    } else {
+		$icce{ $v } = $icc;
+		push @{ $icci{ $icc } }, $v;
+		$icc++;
+	    }
+	}
+	if ($icc) {
+	    @cce{ keys %icce } = values %icce;
+	    @cci{ keys %icci } = values %icci;
+	}
+    } else {
+	my @u = $g->unique_vertices;
+	my %r; @r{ @u } = @u;
+	my $froot = sub {
+	    (each %r)[1];
+	};
+	my $nroot = sub {
+	    $cc++ if keys %r;
+	    (each %r)[1];
+	};
+	my $t = Graph::Traversal::DFS->new($g,
+					   first_root => $froot,
+					   next_root  => $nroot,
+					   pre => sub {
+					       my ($v, $t) = @_;
+					       $cce{ $v } = $cc;
+					       push @{ $cci{ $cc } }, $v;
+					       delete $r{ $v };
+					   },
+					   @_);
+	$t->dfs;
+    }
+    return [ \%cce, \%cci ];
+}
+
 sub _connected_components {
     my $g = shift;
-    my $cce;
-    my $cci;
-    my $ccc = $g->get_graph_attribute('_ccc');
-    if (defined $ccc && $ccc->[ 0 ] == $g->[ _G ]) {
-	($cce, $cci) = @{ $ccc->[ 1 ] };
-    } else {
-	my %cce;
-	my %cci;
-	my $cc = 0;
-	if ($g->has_union_find) {
-	    my $UF = $g->_get_union_find();
-	    my $V  = $g->[ _V ];
-	    my %icce; # Isolated vertices.
-	    my %icci;
-	    my $icc = 0;
-	    for my $v ( $g->unique_vertices ) {
-		$cc = $UF->find( $V->_get_path_id( $v ) );
-		if (defined $cc) {
-		    $cce{ $v } = $cc;
-		    push @{ $cci{ $cc } }, $v;
-		} else {
-		    $icce{ $v } = $icc;
-		    push @{ $icci{ $icc } }, $v;
-		    $icc++;
-		}
-	    }
-	    if ($icc) {
-		@cce{ keys %icce } = values %icce;
-		@cci{ keys %icci } = values %icci;
-	    }
-	} else {
-	    my @u = $g->unique_vertices;
-	    my %r; @r{ @u } = @u;
-	    my $froot = sub {
-		(each %r)[1];
-	    };
-	    my $nroot = sub {
-		$cc++ if keys %r;
-		(each %r)[1];
-	    };
-	    my $t = Graph::Traversal::DFS->new($g,
-					       first_root => $froot,
-					       next_root  => $nroot,
-					       pre => sub {
-						   my ($v, $t) = @_;
-						   $cce{ $v } = $cc;
-						   push @{ $cci{ $cc } }, $v;
-						   delete $r{ $v };
-					       },
-					       @_);
-	    $t->dfs;
-	}
-	($cce, $cci) = (\%cce, \%cci);
-	$g->set_graph_attribute('_ccc', [ $g->[ _G ], [ $cce, $cci ] ]);
-    }
-    return ( $cce, $cci );
+    my $ccc = _check_cache($g, 'connectivity',
+			   \&_connected_components_compute, @_);
+    return @{ $ccc };
 }
 
 sub connected_component_by_vertex {
@@ -2700,53 +2835,54 @@ sub weakly_connected_graph {
     $g->undirected_copy->connected_graph(@_);
 }
 
+sub _strongly_connected_components_compute {
+    my $g = shift;
+    my $t = Graph::Traversal::DFS->new($g);
+    my @d = reverse $t->dfs;
+    my @c;
+    my $h = $g->transpose_graph;
+    my $u =
+	Graph::Traversal::DFS->new($h,
+				   next_root => sub {
+				       my ($t, $u) = @_;
+				       my $root;
+				       while (defined($root = shift @d)) {
+					   last if exists $u->{ $root };
+				       }
+				       if (defined $root) {
+					   push @c, [];
+					   return $root;
+				       } else {
+					   return;
+				       }
+				   },
+				   pre => sub {
+				       my ($v, $t) = @_;
+				       push @{ $c[-1] }, $v;
+				   },
+				   @_);
+    $u->dfs;
+    return \@c;
+}
+
 sub _strongly_connected_components {
     my $g = shift;
-    my $scc = $g->get_graph_attribute('_scc');
-    if (defined $scc && $scc->[ 0 ] == $g->[ _G ]) {
-	$scc = $scc->[ 1 ];
-    } else {
-	my $t = Graph::Traversal::DFS->new($g);
-	my @d = reverse $t->dfs;
-	my @c;
-	my $h = $g->transpose_graph;
-	my $u =
-	    Graph::Traversal::DFS->new($h,
-				       next_root => sub {
-					   my ($t, $u) = @_;
-					   my $root;
-					   while (defined($root = shift @d)) {
-					       last if exists $u->{ $root };
-					   }
-					   if (defined $root) {
-					       push @c, [];
-					       return $root;
-					   } else {
-					       return;
-					   }
-				       },
-				       pre => sub {
-					   my ($v, $t) = @_;
-					   push @{ $c[-1] }, $v;
-				       },
-				       @_);
-	$u->dfs;
-	$scc = \@c;
-	$g->set_graph_attribute('_scc', [ $g->[ _G ], $scc ]);
-    }
-    return @$scc;
+    my $scc = _check_cache($g, 'strong_connectivity',
+			   \&_strongly_connected_components_compute, @_);
+    return defined $scc ? @$scc : ( );
 }
 
 sub strongly_connected_components {
     my $g = shift;
     $g->expect_directed;
-    $g->_strongly_connected_components;
+    $g->_strongly_connected_components(@_);
 }
 
 sub strongly_connected_component_by_vertex {
-    my ($g, $v) = @_;
+    my $g = shift;
+    my $v = shift;
     $g->expect_directed;
-    my @scc = $g->_strongly_connected_components( next_alphabetic => 1);
+    my @scc = $g->_strongly_connected_components( next_alphabetic => 1, @_ );
     for (my $i = 0; $i <= $#scc; $i++) {
 	for (my $j = 0; $j <= $#{ $scc[$i] }; $j++) {
 	    return $i if $scc[$i]->[$j] eq $v;
@@ -2756,16 +2892,17 @@ sub strongly_connected_component_by_vertex {
 }
 
 sub strongly_connected_component_by_index {
-    my ($g, $i) = @_;
+    my $g = shift;
+    my $i = shift;
     $g->expect_directed;
-    my $c = ( $g->_strongly_connected_components )[ $i ];
+    my $c = ( $g->_strongly_connected_components(@_) )[ $i ];
     return defined $c ? @{ $c } : ();
 }
 
 sub same_strongly_connected_components {
     my $g = shift;
     $g->expect_directed;
-    my @scc = $g->_strongly_connected_components( next_alphabetic => 1);
+    my @scc = $g->_strongly_connected_components( next_alphabetic => 1, @_ );
     my @i;
     while (@_) {
 	my $v = shift;
@@ -2918,183 +3055,183 @@ sub _make_bcc {
     return [ values %b, $c ];
 }
 
+sub _biconnectivity_compute {
+    my $g = shift;
+    my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) =
+	$g->_root_opt(@_);
+    return () unless defined $r;
+    my %P;
+    my %I;
+    for my $v ($g->vertices) {
+	$I{ $v } = 0;
+    }
+    $I{ $r } = 1;
+    my %U;
+    my %S; # Self-loops.
+    for my $e ($g->edges) {
+	my ($u, $v) = @$e;
+	$U{ $u }{ $v } = 0;
+	$U{ $v }{ $u } = 0;
+	$S{ $u } = 1 if $u eq $v;
+    }
+    my $i = 1;
+    my $v = $r;
+    my %AP;
+    my %L = ( $r => 1 );
+    my @S = ( $r );
+    my %A;
+    my @V = $g->vertices;
+
+    # print "V : @V\n";
+    # print "r : $r\n";
+
+    my %T; @T{ @V } = @V;
+
+    for my $w (@V) {
+	my @s = $g->successors( $w );
+	if (@s) {
+	    @s = grep { $_ eq $w ? ( delete $T{ $w }, 0 ) : 1 } @s;
+	    @{ $A{ $w } }{ @s } = @s;
+	} elsif ($g->predecessors( $w ) == 0) {
+	    delete $T{ $w };
+	    if ($w eq $r) {
+		delete $I { $r };
+		$r = $v = each %T;
+		if (defined $r) {
+		    %L = ( $r => 1 );
+		    @S = ( $r );
+		    $I{ $r } = 1;
+		    # print "r : $r\n";
+		}
+	    }
+	}
+    }
+
+    # use Data::Dumper;
+    # print "T : ", Dumper(\%T);
+    # print "A : ", Dumper(\%A);
+
+    my %V2BC;
+    my @BR;
+    my @BC;
+
+    my @C;
+    my $Avok;
+
+    while (keys %T) {
+	# print "T = ", Dumper(\%T);
+	do {
+	    my $w;
+	    do {
+		my @w = _shuffle values %{ $A{ $v } };
+		# print "w = @w\n";
+		$w = first { !$U{ $v }{ $_ } } @w;
+		if (defined $w) {
+		    # print "w = $w\n";
+		    $U{ $v }{ $w }++;
+		    $U{ $w }{ $v }++;
+		    if ($I{ $w } == 0) {
+			$P{ $w } = $v;
+			$i++;
+			$I{ $w } = $i;
+			$L{ $w } = $i;
+			push @S, $w;
+			$v = $w;
+		    } else {
+			$L{ $v } = $I{ $w } if $I{ $w } < $L{ $v };
+		    }
+		}
+	    } while (defined $w);
+	    # print "U = ", Dumper(\%U);
+	    # print "P = ", Dumper(\%P);
+	    # print "L = ", Dumper(\%L);
+	    if (!defined $P{ $v }) {
+		# Do nothing.
+	    } elsif ($P{ $v } ne $r) {
+		if ($L{ $v } < $I{ $P{ $v } }) {
+		    $L{ $P{ $v } } = $L{ $v } if $L{ $v } < $L{ $P{ $v } };
+		} else {
+		    $AP{ $P{ $v } } = $P{ $v };
+		    push @C, _make_bcc(\@S, $v, $P{ $v } );
+		}
+	    } else {
+		my $e;
+		for my $w (_shuffle keys %{ $A{ $r } }) {
+		    # print "w = $w\n";
+		    unless ($U{ $r }{ $w }) {
+			$e = $r;
+			# print "e = $e\n";
+			last;
+		    }
+		}
+		$AP{ $e } = $e if defined $e;
+		push @C, _make_bcc(\@S, $v, $r);
+	    }
+	    # print "AP = ", Dumper(\%AP);
+	    # print "C  = ", Dumper(\@C);
+	    # print "L  = ", Dumper(\%L);
+	    $v = defined $P{ $v } ? $P{ $v } : $r;
+	    # print "v = $v\n";
+	    $Avok = 0;
+	    if (defined $v) {
+		if (keys %{ $A{ $v } }) {
+		    if (!exists $P{ $v }) {
+			for my $w (keys %{ $A{ $v } }) {
+			    $Avok++ if $U{ $v }{ $w };
+			}
+			# print "Avok/1 = $Avok\n";
+			$Avok = 0 unless $Avok == keys %{ $A{ $v } };
+			# print "Avok/2 = $Avok\n";
+		    }
+		} else {
+		    $Avok = 1;
+		    # print "Avok/3 = $Avok\n";
+		}
+	    }
+	} until ($Avok);
+
+	last if @C == 0 && !exists $S{$v};
+
+	for (my $i = 0; $i < @C; $i++) {
+	    for my $v (@{ $C[ $i ]}) {
+		$V2BC{ $v }{ $i }++;
+		delete $T{ $v };
+	    }
+	}
+
+	for (my $i = 0; $i < @C; $i++) {
+	    if (@{ $C[ $i ] } == 2) {
+		push @BR, $C[ $i ];
+	    } else {
+		push @BC, $C[ $i ];
+	    }
+	}
+
+	if (keys %T) {
+	    $r = $v = each %T;
+	}
+    }
+    
+    return [ [values %AP], \@BC, \@BR, \%V2BC ];
+}
+
 sub biconnectivity {
     my $g = shift;
     $g->expect_undirected;
-    my $bcc = $g->get_graph_attribute('_bcc');
-    if (defined $bcc && $bcc->[ 0 ] == $g->[ _G ]) {
-	$bcc = $bcc->[ 1 ];
-    } else {
-	my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) =
-	    $g->_root_opt(@_);
-	return () unless defined $r;
-	my %P;
-	my %I;
-	for my $v ($g->vertices) {
-	    $I{ $v } = 0;
-	}
-	$I{ $r } = 1;
-	my %U;
-	for my $e ($g->edges) {
-	    my ($u, $v) = @$e;
-	    $U{ $u }{ $v } = 0;
-	    $U{ $v }{ $u } = 0;
-	}
-	my $i = 1;
-	my $v = $r;
-	my %AP;
-	my %L = ( $r => 1 );
-	my @S = ( $r );
-	my %A;
-	my @V = $g->vertices;
-
-	# print "V : @V\n";
-	# print "r : $r\n";
-
-	my %T; @T{ @V } = @V;
-	my %S; # Self-loops.
-
-	for my $w (@V) {
-	    my @s = $g->successors( $w );
-	    if (@s) {
-		@s = grep { $_ eq $w ? ( delete $T{ $w }, 0 ) : 1 } @s;
-		@{ $A{ $w } }{ @s } = @s;
-	    } elsif ($g->predecessors( $w ) == 0) {
-		delete $T{ $w };
-		if ($w eq $r) {
-		    delete $I { $r };
-		    $r = $v = each %T;
-		    if (defined $r) {
-			%L = ( $r => 1 );
-			@S = ( $r );
-			$I{ $r } = 1;
-			# print "r : $r\n";
-		    }
-		}
-	    }
-	}
-
-	# use Data::Dumper;
-	# print "T : ", Dumper(\%T);
-	# print "A : ", Dumper(\%A);
-
-	my %V2BC;
-	my @BR;
-	my @BC;
-
-	my @C;
-	my $Avok;
-
-	while (keys %T) {
-	    # print "T = ", Dumper(\%T);
-	    do {
-		my $w;
-	        do {
-		    my @w = _shuffle values %{ $A{ $v } };
-		    # print "w = @w\n";
-		    $w = first { !$U{ $v }{ $_ } } @w;
-		    if (defined $w) {
-			# print "w = $w\n";
-			$U{ $v }{ $w }++;
-			$U{ $w }{ $v }++;
-			if ($I{ $w } == 0) {
-			    $P{ $w } = $v;
-			    $i++;
-			    $I{ $w } = $i;
-			    $L{ $w } = $i;
-			    push @S, $w;
-			    $v = $w;
-			} else {
-			    $L{ $v } = $I{ $w } if $I{ $w } < $L{ $v };
-			}
-		    }
-		} while (defined $w);
-		# print "U = ", Dumper(\%U);
-		# print "P = ", Dumper(\%P);
-		# print "L = ", Dumper(\%L);
-		if (!defined $P{ $v }) {
-		    # Do nothing.
-		} elsif ($P{ $v } ne $r) {
-		    if ($L{ $v } < $I{ $P{ $v } }) {
-			$L{ $P{ $v } } = $L{ $v } if $L{ $v } < $L{ $P{ $v } };
-		    } else {
-			$AP{ $P{ $v } } = $P{ $v };
-			push @C, _make_bcc(\@S, $v, $P{ $v } );
-		    }
-		} else {
-		    my $e;
-		    for my $w (_shuffle keys %{ $A{ $r } }) {
-			# print "w = $w\n";
-			unless ($U{ $r }{ $w }) {
-			    $e = $r;
-			    # print "e = $e\n";
-			    last;
-			}
-		    }
-		    $AP{ $e } = $e if defined $e;
-		    push @C, _make_bcc(\@S, $v, $r);
-		}
-		# print "AP = ", Dumper(\%AP);
-		# print "C  = ", Dumper(\@C);
-		# print "L  = ", Dumper(\%L);
-		$v = defined $P{ $v } ? $P{ $v } : $r;
-		# print "v = $v\n";
-		$Avok = 0;
-		if (defined $v) {
-		    if (keys %{ $A{ $v } }) {
-			if (!exists $P{ $v }) {
-			    for my $w (keys %{ $A{ $v } }) {
-				$Avok++ if $U{ $v }{ $w };
-			    }
-			    # print "Avok/1 = $Avok\n";
-			    $Avok = 0 unless $Avok == keys %{ $A{ $v } };
-			    # print "Avok/2 = $Avok\n";
-			}
-		    } else {
-			$Avok = 1;
-			# print "Avok/3 = $Avok\n";
-		    }
-		}
-	    } until ($Avok);
-
-	    last if @C == 0 && !exists $S{$v};
-
-	    for (my $i = 0; $i < @C; $i++) {
-		for my $v (@{ $C[ $i ]}) {
-		    $V2BC{ $v }{ $i }++;
-		    delete $T{ $v };
-		}
-	    }
-
-	    for (my $i = 0; $i < @C; $i++) {
-		if (@{ $C[ $i ] } == 2) {
-		    push @BR, $C[ $i ];
-		} else {
-		    push @BC, $C[ $i ];
-		}
-	    }
-
-	    if (keys %T) {
-		$r = $v = each %T;
-	    }
-	}
-
-	$bcc = [ [values %AP], \@BC, \@BR, \%V2BC ];
-	$g->set_graph_attribute('_bcc', [ $g->[ _G ], $bcc ]);
-    }
-
-    return @$bcc;
+    my $bcc = _check_cache($g, 'biconnectivity',
+			   \&_biconnectivity_compute, @_);
+    return defined $bcc ? @$bcc : ( );
 }
 
 sub is_biconnected {
     my $g = shift;
-    my ($ap, $bc) = ($g->biconnectivity)[0, 1];
+    my ($ap, $bc) = ($g->biconnectivity(@_))[0, 1];
     return defined $ap ? @$ap == 0 && $g->vertices >= 3 : undef;
 }
 
 sub is_edge_connected {
     my $g = shift;
-    my ($br) = ($g->biconnectivity)[2];
+    my ($br) = ($g->biconnectivity(@_))[2];
     return defined $br ? @$br == 0 && $g->edges : undef;
 }
 
@@ -3106,7 +3243,7 @@ sub is_edge_separable {
 
 sub articulation_points {
     my $g = shift;
-    my ($ap) = ($g->biconnectivity)[0];
+    my ($ap) = ($g->biconnectivity(@_))[0];
     return defined $ap ? @$ap : ();
 }
 
@@ -3114,26 +3251,28 @@ sub articulation_points {
 
 sub biconnected_components {
     my $g = shift;
-    my ($bc) = ($g->biconnectivity)[1];
+    my ($bc) = ($g->biconnectivity(@_))[1];
     return defined $bc ? @$bc : ();
 }
 
 sub biconnected_component_by_index {
-    my ($g, $i) = @_;
-    my ($bc) = ($g->biconnectivity)[1];
+    my $g = shift;
+    my $i = shift;
+    my ($bc) = ($g->biconnectivity(@_))[1];
     return defined $bc ? $bc->[ $i ] : undef;
 }
 
 sub biconnected_component_by_vertex {
-    my ($g, $v) = @_;
-    my ($v2bc) = ($g->biconnectivity)[3];
+    my $g = shift;
+    my $v = shift;
+    my ($v2bc) = ($g->biconnectivity(@_))[3];
     return defined $v2bc->{ $v } ? keys %{ $v2bc->{ $v } } : ();
 }
 
 sub same_biconnected_components {
     my $g = shift;
     my $u = shift;
-    my @u = $g->biconnected_component_by_vertex($u);
+    my @u = $g->biconnected_component_by_vertex($u, @_);
     return 0 unless @u;
     my %ubc; @ubc{ @u } = ();
     while (@_) {
@@ -3156,7 +3295,7 @@ sub same_biconnected_components {
 
 sub biconnected_graph {
     my ($g, %opt) = @_;
-    my ($bc, $v2bc) = ($g->biconnectivity)[1, 3];
+    my ($bc, $v2bc) = ($g->biconnectivity, %opt)[1, 3];
     my $bcg = Graph::Undirected->new;
     my $sc_cb =
 	exists $opt{super_component} ?
@@ -3190,7 +3329,7 @@ sub biconnected_graph {
 
 sub bridges {
     my $g = shift;
-    my ($br) = ($g->biconnectivity)[2];
+    my ($br) = ($g->biconnectivity(@_))[2];
     return defined $br ? @$br : ();
 }
 
@@ -3219,6 +3358,9 @@ sub _SPT_add {
     }
 }
 
+sub _SPT_Dijkstra_compute {
+}
+
 sub SPT_Dijkstra {
     my $g = shift;
     my %opt = @_ == 1 ? (first_root => $_[0]) : @_;
@@ -3236,7 +3378,7 @@ sub SPT_Dijkstra {
 
     my $spt = $spt_di->{ $first_root }->[ 1 ];
 
-    $spt->set_graph_attribute('SPT_Dijsktra_root', $first_root);
+    $spt->set_graph_attribute('SPT_Dijkstra_root', $first_root);
 
     return $spt;
 }
@@ -3321,6 +3463,9 @@ sub _SPT_Bellman_Ford {
     return (\%p, \%d);
 }
 
+sub _SPT_Bellman_Ford_compute {
+}
+
 sub SPT_Bellman_Ford {
     my $g = shift;
 
@@ -3361,7 +3506,7 @@ sub SPT_Bellman_Ford {
 
 sub SP_Bellman_Ford {
     my ($g, $u, $v) = @_;
-    my $sptg = $g->SPT_Dijkstra(first_root => $u);
+    my $sptg = $g->SPT_Bellman_Ford(first_root => $u);
     my @path = ($v);
     my %seen;
     my $V = $g->vertices;
@@ -3399,6 +3544,9 @@ sub APSP_Floyd_Warshall {
 
 *all_pairs_shortest_paths = \&APSP_Floyd_Warshall;
 
+sub _transitive_closure_matrix_compute {
+}
+
 sub transitive_closure_matrix {
     my $g = shift;
     my $tcm = $g->get_graph_attribute('_tcm');
@@ -3533,19 +3681,25 @@ sub longest_path {
 
 sub vertex_eccentricity {
     my ($g, $u) = @_;
-    my $max;
-    for my $v ($g->vertices) {
-	next if $u eq $v;
-	my $l = $g->path_length($u, $v);
-	if (defined $l && (!defined $max || $l > $max)) {
-	    $max = $l;
+    $g->expect_undirected;
+    if ($g->is_connected) {
+	my $max;
+	for my $v ($g->vertices) {
+	    next if $u eq $v;
+	    my $l = $g->path_length($u, $v);
+	    if (defined $l && (!defined $max || $l > $max)) {
+		$max = $l;
+	    }
 	}
+	return $max;
+    } else {
+	return Infinity();
     }
-    return $max;
 }
 
 sub shortest_path {
     my ($g, $u, $v) = @_;
+    $g->expect_undirected;
     my $t = $g->transitive_closure_matrix;
     if (defined $u) {
 	if (defined $v) {
@@ -3586,12 +3740,18 @@ sub shortest_path {
 
 sub radius {
     my $g = shift;
-    my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
-    return defined $minp ? (wantarray ? @$minp : $min) : undef;
+    $g->expect_undirected;
+    my ($center, $radius) = (undef, Infinity());
+    for my $v ($g->vertices) {
+	my $x = $g->vertex_eccentricity($v);
+	($center, $radius) = ($v, $x) if defined $x && $x < $radius;
+    }
+    return $radius;
 }
 
 sub center_vertices {
     my ($g, $delta) = @_;
+    $g->expect_undirected;
     $delta = 0 unless defined $delta;
     $delta = abs($delta);
     my @c;
@@ -3721,7 +3881,6 @@ sub could_be_isomorphic {
 		       $d0{$da}{$db} == $d1{$da}{$db};
 	}
     }
-    my $f = 1;
     for my $da (keys %d0) {
 	for my $db (keys %{ $d0{$da} }) {
 	    return 0 unless $d1{$da}{$db} == $d0{$da}{$db};
@@ -3729,14 +3888,138 @@ sub could_be_isomorphic {
 	delete $d1{$da};
     }
     return 0 unless keys %d1 == 0;
+    my $f = 1;
     for my $da (keys %d0) {
 	for my $db (keys %{ $d0{$da} }) {
-	    $f *= _factorial(abs($d0{$da}{$db})) if $d0{$da}{$db} > 1;
+	    $f *= _factorial(abs($d0{$da}{$db}));
 	}
     }
     return $f;
 }
 
+###
+# Analysis functions.
+
+sub subgraph_by_radius
+{
+    my ($g, $n, $rad) = @_;
+
+    return unless defined $n && defined $rad && $rad >= 0;
+
+    my $r = (ref $g)->new;
+
+    if ($rad == 0) {
+	return $r->add_vertex($n);
+    }
+
+    my %h;
+    $h{1} = [ [ $n, $g->successors($n) ] ];
+    for my $i (1..$rad) {
+	$h{$i+1} = [];
+	for my $arr (@{ $h{$i} }) {
+	    my ($p, @succ) = @{ $arr };
+	    for my $s (@succ) {
+		$r->add_edge($p, $s);
+		push(@{ $h{$i+1} }, [$s, $g->successors($s)]) if $i < $rad;
+	    }
+	}
+    }
+
+    return $r;
+}
+
+sub clustering_coefficient {
+    my ($g) = @_;
+    my %clustering;
+
+    my $gamma = 0;
+
+    for my $n ($g->vertices()) {
+	my $gamma_v = 0;
+	my @neigh = $g->successors($n);
+	my %c;
+	for my $u (@neigh) {
+	    for my $v (@neigh) {
+		if (!$c{"$u-$v"} && $g->has_edge($u, $v)) {
+		    $gamma_v++;
+		    $c{"$u-$v"} = 1;
+		    $c{"$v-$u"} = 1;
+		}
+	    }
+	}
+	if (@neigh > 1) {
+	    $clustering{$n} = $gamma_v/(@neigh * (@neigh - 1) / 2);
+	    $gamma += $gamma_v/(@neigh * (@neigh - 1) / 2);
+	} else {
+	    $clustering{$n} = 0;
+	}
+    }
+
+    $gamma /= $g->vertices();
+
+    return wantarray ? ($gamma, %clustering) : $gamma;
+}
+
+sub betweenness {
+    my $g = shift;
+
+    my @V = $g->vertices();
+
+    my %Cb; # C_b{w} = 0
+
+    $Cb{$_} = 0 for @V;
+
+    for my $s (@V) {
+	my @S; # stack (unshift, shift)
+
+	my %P; # P{w} = empty list
+	$P{$_} = [] for @V;
+
+	my %sigma; # \sigma{t} = 0
+	$sigma{$_} = 0 for @V;
+	$sigma{$s} = 1;
+
+	my %d; # d{t} = -1;
+	$d{$_} = -1 for @V;
+	$d{$s} = 0;
+
+	my @Q; # queue (push, shift)
+	push @Q, $s;
+
+	while (@Q) {
+	    my $v = shift @Q;
+	    unshift @S, $v;
+	    for my $w ($g->successors($v)) {
+		# w found for first time
+		if ($d{$w} < 0) {
+		    push @Q, $w;
+		    $d{$w} = $d{$v} + 1;
+		}
+		# Shortest path to w via v
+		if ($d{$w} == $d{$v} + 1) {
+		    $sigma{$w} += $sigma{$v};
+		    push @{ $P{$w} }, $v;
+		}
+	    }
+	}
+
+	my %delta;
+	$delta{$_} = 0 for @V;
+
+	while (@S) {
+	    my $w = shift @S;
+	    for my $v (@{ $P{$w} }) {
+		$delta{$v} += $sigma{$v}/$sigma{$w} * (1 + $delta{$w});
+	    }
+	    if ($w ne $s) {
+		$Cb{$w} += $delta{$w};
+	    }
+	}
+    }
+
+    return %Cb;
+}
+
 ###
 # Debugging.
 #
